import sys
import numpy as np
from nltk.tokenize import word_tokenize
from scipy.sparse import hstack
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import confusion_matrix
import json
import ast
from sklearn.metrics import f1_score,classification_report
from copy import deepcopy
from sklearn.svm import LinearSVC
from data_config import DAT

def create_data(DATA):
    train_words = []
    dev_words = []
    test_words = []
    
    train_labels = []
    dev_labels = []
    test_labels = []
    if "dev" in DAT[DATA]:
        PTRAIN = DAT[DATA]["train"]
        PDEV = DAT[DATA]["dev"]
        PTEST = DAT[DATA]["test"]

    if "dev" in DAT[DATA]:
        with open(PTRAIN,encoding="utf-8") as f:
            lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
            for i,l in enumerate(lines):
                sw = word_tokenize(l[1])
                tw = word_tokenize(l[2])
                bw = [t for t in sw if t in tw]
                train_labels.append(l[-1])
                train_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
                #print(l[5])
                """"
                paths = ast.literal_eval(l[5])
                for j,pa in enumerate(paths):
                    pw = pa.split()
                    train_words[i]+=["PATH"+str(j)+":" + t for t in pw]
                """
                if i %100 == 0:
                    print(i,"train examples created")
                
        with open(PDEV,encoding="utf-8") as f:
            lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
            for i,l in enumerate(lines):
                sw = word_tokenize(l[1])
                tw = word_tokenize(l[2])
                bw = [t for t in sw if t in tw]
                dev_labels.append(l[-1])
                dev_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
                """
                paths = ast.literal_eval(l[5])
                for j,pa in enumerate(paths):
                    pw =pa.split()
                    dev_words[i]+=["PATH"+str(j)+":" + t for t in pw]
                """
    
    #just in case you don't have dev set, the we split of dev from train
    else:
        attacks=0
        supports=0
        trainidx=0
        devidx=0
        with open(PTRAIN) as f:
            lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
            for i,l in enumerate(lines):
                sw = word_tokenize(l[1])
                tw = word_tokenize(l[2])
                bw = [t for t in sw if t in tw]
                if l[-1] == "support" and supports <=46:
                    supports+=1
                    dev_labels.append(l[-1])
                    dev_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in sw if t in tw])
                    """
                    paths = ast.literal_eval(l[5])
                    for j,pa in enumerate(paths):
                        pw = pa.split()
                        dev_words[devidx]+=["PATH"+str(j)+":" + t for t in pw]
                    """
                    devidx+=1
                elif l[-1] == "attack" and attacks <= 20:
                    attacks+=1
                    dev_labels.append(l[-1])
                    dev_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in sw if t in tw])
                    """
                    paths = ast.literal_eval(l[5])
                    for j,pa in enumerate(paths):
                        pw = pa.split()
                        dev_words[devidx]+=["PATH"+str(j)+":" + t for t in pw]
                    """
                    devidx+=1

                else:
                    train_labels.append(l[-1])
                    train_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in sw if t in tw])
                    """
                    paths = ast.literal_eval(l[5])
                    for j,pa in enumerate(paths):
                        pw = pa.split()
                        train_words[trainidx]+=["PATH"+str(j)+":" + t for t in pw]
                    """
                    trainidx+=1

    with open(PTEST,encoding="utf-8") as f:
        lines = [l.split("\t") for l in f.read().split("\n") if "\t"  in l]
        for i,l in enumerate(lines):
            sw = word_tokenize(l[1])
            tw = word_tokenize(l[2])
            bw = [t for t in sw if t in tw]
            test_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
            test_labels.append(l[-1])
            """
            paths = ast.literal_eval(l[5])
            for j,pa in enumerate(paths):
                pw = pa.split()
                test_words[i]+=["PATH"+str(j)+":" + t for t in pw]
            """
    for i,la in enumerate(train_labels):
        if "support" in la:
            train_labels[i] = "support"
        elif "attack" in la:
            train_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    for i,la in enumerate(dev_labels):
        if "support" in la:
            dev_labels[i] = "support"
        elif "attack" in la:
            dev_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    for i,la in enumerate(test_labels):
        if "support" in la:
            test_labels[i] = "support"
        elif "attack" in la:
            test_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    return train_words,dev_words,test_words,train_labels,dev_labels,test_labels

def get_context(PC=""):
    dic={}
    with open(PC,"r",encoding="utf-8") as f:
        lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
        for l in lines:
            key = l[0].split("_",1)[1]
            dic[key] = {"cs":word_tokenize(l[1])}
            dic[key]["ct"] = word_tokenize(l[2])
    return dic

def create_data_context(DATA):
    train_words = []
    dev_words = []
    test_words = []
    
    train_context = []
    dev_context = []
    test_context = []
    
    train_labels = []
    dev_labels = []
    test_labels = []
    stuc = get_context(DAT[DATA]["ctx"])
    if "dev" in DAT[DATA]:
        PTRAIN = DAT[DATA]["train"]
        PDEV = DAT[DATA]["dev"]
        PTEST = DAT[DATA]["test"]
    with open(PTRAIN,encoding="utf-8") as f:
        lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
        for i,l in enumerate(lines):
            sw = word_tokenize(l[1])
            tw = word_tokenize(l[2])
            bw = [t for t in sw if t in tw]
            train_labels.append(l[-1])
            train_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
            train_context.append(["SOURCE:"+t for t in stuc[l[0]]["cs"]]+["TARGET:"+t for t in stuc[l[0]]["ct"]])
            """
            paths = ast.literal_eval(l[5])
            for j,pa in enumerate(paths):
                pw = pa.split()
                train_words[i]+=["PATH"+str(j)+":" + t for t in pw]
            """
            if i %100 == 0:
                print(i,"train examples created")
    
    with open(PDEV,encoding="utf-8") as f:
        lines = [l.split("\t") for l in f.read().split("\n") if "\t" in l]
        for i,l in enumerate(lines):
            sw = word_tokenize(l[1])
            tw = word_tokenize(l[2])
            bw = [t for t in sw if t in tw]
            dev_labels.append(l[-1])
            dev_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
            dev_context.append(["SOURCE:"+t for t in stuc[l[0]]["cs"]]+["TARGET:"+t for t in stuc[l[0]]["ct"]])
            """
            paths = ast.literal_eval(l[5])
            for j,pa in enumerate(paths):
                pw =pa.split()
                dev_words[i]+=["PATH"+str(j)+":" + t for t in pw]
            """

    with open(PTEST,encoding="utf-8") as f:
        lines = [l.split("\t") for l in f.read().split("\n") if "\t"  in l]
        for i,l in enumerate(lines):
            idx = l[0]
            sw = word_tokenize(l[1])
            tw = word_tokenize(l[2])
            bw = [t for t in sw if t in tw]
            test_words.append(["SOURCE:"+t for t in sw]+["TARGET:"+t for t in tw]+["BOTH:"+t for t in bw])
            test_labels.append(l[-1])
            test_context.append(["SOURCE:"+t for t in stuc[l[0]]["cs"]]+["TARGET:"+t for t in stuc[l[0]]["ct"]])
            """
            paths = ast.literal_eval(l[5])
            for j,pa in enumerate(paths):
                pw = pa.split()
                test_words[i]+=["PATH"+str(j)+":" + t for t in pw]
            """
    for i,la in enumerate(train_labels):
        if "support" in la:
            train_labels[i] = "support"
        elif "attack" in la:
            train_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    for i,la in enumerate(dev_labels):
        if "support" in la:
            dev_labels[i] = "support"
        elif "attack" in la:
            dev_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    for i,la in enumerate(test_labels):
        if "support" in la:
            test_labels[i] = "support"
        elif "attack" in la:
            test_labels[i] = "attack"
        else:
            print("forbidden label",la)
            sys.exit(1)
    return train_words,dev_words,test_words,train_labels,dev_labels,test_labels,train_context,dev_context,test_context

