import numpy as np

from sklearn.metrics import roc_auc_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import classification_report
from sklearn.metrics import f1_score

def predict_nli(model, X):
    _, p_embedding, n_embedding= model.predict(X)
    
    scores_p = np.einsum("ij -> i", p_embedding)
    scores_n = np.einsum("ij -> i", n_embedding)
    return np.vstack((scores_p, scores_n))


def get_classif_report(model,X,labels,bin_labels=["support","attack"]):
    if type(model) != list:
        scores = predict_nli(model,X).T
    else:
        scores = predict_nli(model[0],X).T

        for m in model[1:]:
            print(scores[:10])
            scores+=predict_nli(m,X).T
    nodecision = [i for i in range(len(scores)) if abs(scores[i][0] - scores[i][1]) < 0.00001]
    print("nodec",len(nodecision),set(labels))
    correct = [i for i in range(len(scores)) if scores[i][0] > scores[i][1] and i not in nodecision]
    incorrect = [i for i in range(len(scores)) if scores[i][0] < scores[i][1] and i not in nodecision]
    preds = []
    for i,la in enumerate(labels):
        if i in correct:
            preds.append(labels[i])
        else:
            preds.append([ l for l in bin_labels if l != la][0])
    print(classification_report(labels,preds))
    return f1_score(labels,preds,average="macro")


#this is the functio we use
def get_classif_report_scores(scores,X=None,labels=None,bin_labels=["support","attack"],printreport=False,evalfun=f1_score,pos_label=None):
    nodecision = [i for i in range(len(scores)) if abs(scores[i][0] - scores[i][1]) < 0.00001]
    
    #correct are examples where scores[0] > scores[1] 
    #(two instances processed by same simaese network, 
    #the correct one is the first, goal: first achieves higher score than second)
    correct = [i for i in range(len(scores)) if scores[i][0] > scores[i][1] and i not in nodecision]
    incorrect = [i for i in range(len(scores)) if scores[i][0] < scores[i][1] and i not in nodecision]
    preds = []
    for i,la in enumerate(labels):
        # if correct label achieves higher score take true label
        if i in correct:
            preds.append(labels[i])
        # if false label achieves higher score take opposite label
        else:
            preds.append([ l for l in bin_labels if l != la][0])
    if printreport:
        print(classification_report(labels,preds))
    if pos_label:
        return evalfun(labels,preds,pos_label=pos_label)

    return evalfun(labels,preds,average="macro")
