from __future__ import print_function
import os
import sys
import numpy as np
from random import shuffle
from flair.embeddings import ELMoEmbeddings, Sentence, BertEmbeddings 
import json
import data_helpers as dh


def string2file(string,fp):
    with open(fp,"w") as f:
        f.write(string)


#joins tokens
def jt(tokenized_sents):
    return [" ".join(s) for s in tokenized_sents]


def maybe_add_fullstop(xss):
    for xs in xss:
        if xs[-1] not in [".",",","?",";","."]:
            xs.append(".")
    return None

def maybe_target_c(ls,i):
    if not any(l for l in ls):
        return ""
    elm = [t.replace("TARGET:","") for t in ls[i] if "TARGET:" in t]
    elm = [t for t in elm if t != "."]
    elm = " ".join(elm).split(",")[0].split()
    if not elm:
        return ""
    else:
        #return ""
        return " ".join(elm)




def decide_connection(label,poslabel=["because"],neglabel=["however"],switch=False,context=[]):
    # this is needed for constructing training examples -- if label is support, we insert e.g. moreover
    #if label is  attack we e.g., insert however
    
    scontext = [t.replace("SOURCE:","") for t in context if "SOURCE:" in t]
    scontext = [t for t in scontext if t != "."]
    scontext = " ".join(scontext).split(",")[0].split()
    print(scontext)
    if label in ["support","supports"]:
        if switch:
            return scontext+neglabel
        return scontext+poslabel
    elif label in ["attack","attacks"]:
        if switch:
            return scontext+poslabel
        return scontext+neglabel
    else:
        print("error, forbidden label:",label)
        sys.exit(1)
    
def prepare_markers(markss,length):
    deliss = []
    for xs in markss:
        while len(xs) < length:
            xs.append(1)
        deli=[]
        while len(xs) > length:
            if xs[0] == 2:
                del xs[0]
                deli.append(0)
            else:
                del xs[-1]
                deli.append(-1)
        deliss.append(deli)
    return None

def sumlayers(x,path=""):
    #print(x.shape)
    #glo = x[:glove]
    #rest=x[glove:]
    if "bert" in path:
        out=x.reshape(4,1024).sum(axis=0)
    elif "elmo" in path:
        out=x.reshape(4,768).sum(axis=0)
    else:
        print(path)
        print("unknown type of word embedding")
        sys.exit(1)
    return out

def zeropad(xss,path="",length=50,delids=None):
    
    if "bert" in path:
        dim=1024
    elif "elmo" in path:
        dim =768
    else:
        print(path)
        print("unknown type of word embedding")
        sys.exit(1)
    for j,xs in enumerate(xss):
        while len(xs) < length:
            xs.append(np.zeros(dim))
        if not delids:
            while len(xs) > length:
                del xs[0]
        else:
            for i in delids[j]:
                del xs[i]
    return None

def sum_and_save(tokenss,path,length=50,delids=None):
    #we save the vecs so we can use it later again without having to wait long for their inference
    vecs=[[sumlayers(t.embedding.numpy(),path=path) for t in tokens] for tokens in tokenss]
    zeropad(vecs,path=path,length=length,delids=delids)
    vecs=np.array(vecs)

    np.save(path,vecs)
    return vecs


class PreProcessor():

    def __init__(self,DATA=None,EMB=None,context=None,sup_connect="Moreover+,",att_connect="However+,",RUNID=None,maxlen=None):
        self.DATA = DATA
        self.EMB=EMB
        self.att_connect=att_connect
        self.sup_connect=sup_connect
        self.context=context
        self.RUNID = RUNID
        self.maxlen=maxlen
    
    def get_preprocessed_data(self):
        
        if self.context:
            traintoks,devtoks,testtoks,trainlabels,devlabels,testlabels,traincontext,devcontext,testcontext =dh.create_data_context(self.DATA)
        else:
            traintoks,devtoks,testtoks,trainlabels,devlabels,testlabels =dh.create_data(self.DATA)
            traincontext,devcontext,testcontext = [[]]*len(traintoks),[[]]*len(devtoks),[[]]*len(testtoks)

        train_toks_s = [[t.replace("SOURCE:","") for t in ts if "SOURCE:" in t] for ts in traintoks]
        dev_toks_s = [[t.replace("SOURCE:","") for t in ts if "SOURCE:" in t] for ts in devtoks]
        test_toks_s = [[t.replace("SOURCE:","") for t in ts if "SOURCE:" in t] for ts in testtoks]

        train_toks_t = [[t.replace("TARGET:","") for t in ts if "TARGET:" in t] for ts in traintoks]
        dev_toks_t = [[t.replace("TARGET:","") for t in ts if "TARGET:" in t] for ts in devtoks]
        test_toks_t = [[t.replace("TARGET:","") for t in ts if "TARGET:" in t] for ts in testtoks]


        maybe_add_fullstop(train_toks_t)
        maybe_add_fullstop(dev_toks_t)
        maybe_add_fullstop(test_toks_t)

            
        # I+agree --> [I, agree]
        pl =self.sup_connect.replace("+"," ").split()
        nl =self.att_connect.replace("+"," ").split()
        

        # here we generate the markers... somewhat ugly.
        #    [tgt rel token 1,... tgt rel token n, connector token1 ... connector token k, filler, src rel token 1 ... src rel token m]
        # -->[       4        ...          4           3                        3             1          5                   5
        train_marks=[ [2]*(len(maybe_target_c(traincontext,i).split())+1) 
                       +[4]*len(train_toks_t[i])
                       +[3]*(len(decide_connection(trainlabels[i],pl,nl,switch=False,context=traincontext[i]))-2)
                       +[1]*2   
                       +[5]*len(train_toks_s[i]) for i in range(len(train_toks_s))
                       ]  

        dev_marks=[ [2]*(len(maybe_target_c(devcontext,i).split())+1) 
                     +[4]*len(dev_toks_t[i])
                     + [3]*(len(decide_connection(devlabels[i],pl,nl,switch=False,context=devcontext[i]))-2) 
                     + [1]*2   
                     +[5]*len(dev_toks_s[i]) for i in range(len(dev_toks_s))
                  ]  

        test_marks=[ [2]*(len(maybe_target_c(testcontext,i).split())+1) 
                     + [4]*len(test_toks_t[i])
                     + [3]*(len(decide_connection(testlabels[i],pl,nl,switch=False,context=testcontext[i]))-2) 
                     + [1]*2   
                     +[5]*len(test_toks_s[i]) for i in range(len(test_toks_s))
                ]  
        

        #padding and cropping
        delidstrain= prepare_markers(train_marks,self.maxlen)
        print(train_marks[:15])
        delidsdev=prepare_markers(dev_marks,self.maxlen)
        delidstest=prepare_markers(test_marks,self.maxlen)

        print(train_marks[:15])
        train_marks = np.array(train_marks)
        dev_marks = np.array(dev_marks)
        test_marks = np.array(test_marks)
        
        #generate positive (plausible) instance for train dev test
        train_toks_pos = [maybe_target_c(traincontext,i)+" "+" ".join(train_toks_t[i] + decide_connection(trainlabels[i],pl,nl,switch=False,context=traincontext[i]) + train_toks_s[i]).lower() for i in range(len(train_toks_t))]
        print(train_toks_pos[:10])
        print(traincontext[:10])

        dev_toks_pos = [maybe_target_c(devcontext,i)+" "+" ".join(dev_toks_t[i] + decide_connection(devlabels[i],pl,nl,switch=False,context=devcontext[i]) + dev_toks_s[i]).lower() for i in range(len(dev_toks_t))]
        test_toks_pos = [maybe_target_c(testcontext,i)+" "+" ".join(test_toks_t[i] + decide_connection(testlabels[i],pl,nl,switch=False,context=testcontext[i]) + test_toks_s[i]).lower() for i in range(len(test_toks_t))]



        #generate negative (implausible) instances for train dev test
        train_toks_neg = [maybe_target_c(traincontext,i)+" "+" ".join(train_toks_t[i] + decide_connection(trainlabels[i],pl,nl,switch=True,context=traincontext[i]) + train_toks_s[i]).lower() for i in range(len(train_toks_t))]
        dev_toks_neg = [maybe_target_c(devcontext,i)+" "+" ".join(dev_toks_t[i] + decide_connection(devlabels[i],pl,nl,switch=True,context=devcontext[i]) + dev_toks_s[i]).lower() for i in range(len(dev_toks_t))]
        test_toks_neg = [maybe_target_c(testcontext,i)+" "+" ".join(test_toks_t[i] + decide_connection(testlabels[i],pl,nl,switch=True,context=testcontext[i]) + test_toks_s[i]).lower() for i in range(len(test_toks_t))]

        print(list(zip(train_toks_pos[:5],train_toks_neg[:5])))
        print(list(zip(dev_toks_pos[:5],dev_toks_neg[:5])))
        print(list(zip(test_toks_pos[:5],test_toks_neg[:5])))
        print(sorted([len(x.split()) for x in train_toks_pos]))
        


        #infer contextual embeddings and save em or load them if they are already on our disk
        if self.EMB == "elmo":
            embedding = ELMoEmbeddings()
        elif self.EMB == "bert":
            embedding=BertEmbeddings('bert-large-uncased')

        if os.path.isfile(self.DATA+"-"+self.EMB+"train-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            trainxwords_pos=np.load(self.DATA+"-"+self.EMB+"train-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            trainxwords_pos = [Sentence(x,use_tokenizer=False) for x in train_toks_pos]
            print("trainsents created")
            [embedding.embed(s) for s in trainxwords_pos]
            #print(trainxwords[0][0].embedding.numpy().shape)
            print("trainvecs created")
            trainxwords_pos = sum_and_save(trainxwords_pos,self.DATA+"-"+self.EMB+"train-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidstrain)

        if os.path.isfile(self.DATA+"-"+self.EMB+"dev-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            devxwords_pos=np.load(self.DATA+"-"+self.EMB+"dev-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            devxwords_pos = [Sentence(x,use_tokenizer=False) for x in dev_toks_pos]
            print([len(x) for x in devxwords_pos])
            print("devsents created")
            [embedding.embed(s) for s in devxwords_pos]
            print("devvecs created")
            devxwords_pos = sum_and_save(devxwords_pos,self.DATA+"-"+self.EMB+"dev-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidsdev)

        if os.path.isfile(self.DATA+"-"+self.EMB+"test-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            testxwords_pos=np.load(self.DATA+"-"+self.EMB+"test-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            testxwords_pos = [Sentence(x,use_tokenizer=False) for x in test_toks_pos]
            print([len(x) for x in testxwords_pos])
            print("testsents created")
            [embedding.embed(s) for s in testxwords_pos]
            print("testvecs created")
            testxwords_pos = sum_and_save(testxwords_pos,self.DATA+"-"+self.EMB+"test-pos-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidstest)



        if os.path.isfile(self.DATA+"-"+self.EMB+"train-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            trainxwords_neg=np.load(self.DATA+"-"+self.EMB+"train-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            trainxwords_neg = [Sentence(x,use_tokenizer=False) for x in train_toks_neg]
            print([len(x) for x in trainxwords_neg])
            print("trainsents created")
            [embedding.embed(s) for s in trainxwords_neg]
            print("trainvecs created")
            trainxwords_neg = sum_and_save(trainxwords_neg,self.DATA+"-"+self.EMB+"train-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidstrain)

        if os.path.isfile(self.DATA+"-"+self.EMB+"dev-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            devxwords_neg=np.load(self.DATA+"-"+self.EMB+"dev-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            devxwords_neg = [Sentence(x,use_tokenizer=False) for x in dev_toks_neg]
            print([len(x) for x in devxwords_neg])
            print("devsents created")
            [embedding.embed(s) for s in devxwords_neg]
            print("devvecs created")
            devxwords_neg = sum_and_save(devxwords_neg,self.DATA+"-"+self.EMB+"dev-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidsdev)

        if os.path.isfile(self.DATA+"-"+self.EMB+"test-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy"):
            testxwords_neg=np.load(self.DATA+"-"+self.EMB+"test-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy")
        else:
            testxwords_neg = [Sentence(x,use_tokenizer=False) for x in test_toks_neg]
            print([len(x) for x in testxwords_neg])
            print("testsents created")
            [embedding.embed(s) for s in testxwords_neg]
            print("testvecs created")
            testxwords_neg = sum_and_save(testxwords_neg,self.DATA+"-"+self.EMB+"test-neg-CONTEXT="+str(self.context)+"-"+self.sup_connect+"-"+self.att_connect+".npy",self.maxlen,delidstest)


        # Finally we have the data to be fed thorugh our model
        X = {
            'positive_input': trainxwords_pos,
            'negative_input': trainxwords_neg,
            'input_markers': train_marks,
        }

        valX = {
            'positive_input': devxwords_pos,
            'negative_input': devxwords_neg,
            'input_markers': dev_marks,
        }


        testX = {
            'positive_input': testxwords_pos,
            'negative_input': testxwords_neg,
            'input_markers':test_marks,
        }
        return X,valX,testX,trainlabels,devlabels,testlabels


