import metrics
import ast 
import numpy as np
import os
import pprint
from scipy.stats import logistic
from sklearn import metrics as skmetrics

##### evaluate predictions ##########

def fread(p):
    with open(p) as f:
        return f.read()


# we collect the resultsover five random runs to later average over them and compute st
macrof1s= {}
macrof1s_avg= {}

preds = {}

evalfun= skmetrics.f1_score
pos_label=None

def normalize_preds(dn2array):
    # normalize two plausibility scores by selu with the max of the two values
    return np.array([x/max(x) for x in dn2array])



for fi in [x for x in os.listdir("logs") if "ESSAY95" not in x]:
    string = fread("logs/"+fi)
    # the preds of  the model which performed best on dev
    best_ckpt_preds = np.array(ast.literal_eval(string.split("best on dev, test predictions: ")[1].split("\n")[0]))

    #the preds of  the 5 models which performed best on dev (not used in paper)
    best_ckpts_preds = np.array(ast.literal_eval(string.split("best 5 on dev, test predictions: ")[1].split("\n")[0]))
    
    #normalize preds by max (per data example prediction), 
    #e.g. plausibility prediction for data example i: [1.2,1.6] ---> normalized to ---> [1.2/1.6 , 1])
    print(best_ckpt_preds)
    best_ckpt_preds = normalize_preds(best_ckpt_preds)
    print(best_ckpt_preds)
    best_ckpts_preds = np.array([normalize_preds(e) for i,e in enumerate(best_ckpts_preds)])
    
    # we ensemble average the results of the last five cpts (again, not used in paper)
    preds[fi]={"ckpt-avg":np.einsum("ijk -> jk",best_ckpts_preds),"best-dev":best_ckpt_preds}
    
    # get reference
    labs = ast.literal_eval(string.split("ground truth: ")[1].split("\n")[0])
    
    #compute metrics
    a=metrics.get_classif_report_scores(best_ckpt_preds,labels=labs,evalfun=evalfun,pos_label=pos_label)
    b=metrics.get_classif_report_scores(np.einsum("ijk -> jk",best_ckpts_preds),labels=labs,evalfun=evalfun,pos_label=pos_label)
    macrof1s[fi] = a#metrics.get_classif_report_scores(best_ckpt_preds,labels=labs)
    macrof1s[fi+"ckpt-avg"] =b# metrics.get_classif_report_scores(np.einsum("ijk -> jk",best_ckpts_preds),labels=labs)
    runid = fi.split("runid=")[1].split("-")[0]
    fi_glob = fi.replace("runid="+runid,"")
    if fi_glob not in macrof1s_avg:
        macrof1s_avg[fi_glob]=[]
        macrof1s_avg[fi_glob+"ckpt-avg"]=[]
    macrof1s_avg[fi_glob].append(a)
    macrof1s_avg[fi_glob+"ckpt-avg"].append(b)
    
    #if we have collected all 5 runs from the different random inits we can compute mean results and std
    if len(macrof1s_avg[fi_glob]) == 5:
        macrof1s_avg[fi_glob] = [np.mean(macrof1s_avg[fi_glob]),np.std(macrof1s_avg[fi_glob]),macrof1s_avg[fi_glob]]
        macrof1s_avg[fi_glob+"ckpt-avg"] = [np.mean(macrof1s_avg[fi_glob+"ckpt-avg"]),np.std(macrof1s_avg[fi_glob+"ckpt-avg"]),macrof1s_avg[fi_glob+"ckpt-avg"]]
    print(macrof1s[fi]>macrof1s[fi+"ckpt-avg"])
print("finished")
pprint.pprint(macrof1s_avg)

print("\n\n\nMAIN RESULT\n--------------\nFirst number: mean macro F1\nSecond number: std\nFollowing list: input results of independent runs\n")
for key in sorted(list(macrof1s_avg.keys())):
    print(key,macrof1s_avg[key])
print("\n--------Finish----------\n")


# another ensemble option (this is used in paper), we ensemble preds bert over connectors, avg over random seeds
x = []
xcpt=[]
for i in range(0,5):
    tmp1=[]
    tmp2=[]
    for fi in preds:
        runid = str(i)
        tmpid = fi.split("runid=")[1].split("-")[0]
        if runid == tmpid and  "bert" in fi and "context=False" in fi and "markers=True" in fi and "ESSAY95" not in fi and "sup_connect=+" not in fi:
            tmp1.append(preds[fi]["best-dev"])
            tmp2.append(preds[fi]["ckpt-avg"])
    ens1=np.einsum("ijk -> jk",tmp1)
    ens2=np.einsum("ijk -> jk",tmp2)
    x.append(metrics.get_classif_report_scores(ens1,labels=labs, evalfun=evalfun,pos_label=pos_label))
    xcpt.append(metrics.get_classif_report_scores(ens2,labels=labs, evalfun=evalfun,pos_label=pos_label))

print("Result ensemble over connectors",np.mean(x),np.std(x),x)
print("Result ensemble over connctors and ckpts", np.mean(xcpt),np.std(xcpt),xcpt)




    


